import os
import json
import time
from typing import Any, Dict, Optional

import hashlib
import requests
import zipfile

from seaplane_framework.api.exceptions import ApiException

from seaplane.config import config
from seaplane.kv import KeyValueStorageAPI
from seaplane.logs import log
from seaplane.object import ObjectStorageAPI
from seaplane.sdk_internal_utils.http import headers
from seaplane.sdk_internal_utils.token_auth import with_token


def _file_md5(path: str) -> str:
    """
    Gets the MD5 hash of a file by path.
    """

    hasher = hashlib.md5()
    block_size = 4194304  # 4 MB
    with open(path, "rb") as fh:
        while True:
            buffer = fh.read(block_size)
            if not buffer:
                break
            hasher.update(buffer)
    return hasher.hexdigest()


@with_token
def create_stream(token: str, name: str) -> None:
    log.logger.debug(f"creating stream: {name}")
    url = f"{config.carrier_endpoint}/stream/{name}"

    payload: Dict[str, Any] = {"ack_timeout": 20}  # should be long enough for OpenAI
    if config.region is not None:
        payload["allow_locations"] = [f"region/{config.region}"]
    resp = requests.put(
        url,
        json=payload,
        headers=headers(token),
    )
    resp.raise_for_status()


@with_token
def delete_stream(token: str, name: str) -> None:
    log.logger.debug(f"deleting stream: {name}")
    url = f"{config.carrier_endpoint}/stream/{name}"

    resp = requests.delete(
        url,
        headers=headers(token),
    )
    resp.raise_for_status()


@with_token
def add_secrets(token: str, name: str, secrets: Dict[str, str]) -> None:
    log.logger.debug(f"adding secrets: {name}")
    url = f"{config.carrier_endpoint}/flow/{name}/secrets"
    if config.dc_region is not None:
        url += f"?region={config.dc_region}"

    flow_secrets = {}
    for secret_key, secret_value in secrets.items():
        flow_secrets[secret_key] = {"destination": "all", "value": secret_value}

    resp = requests.put(
        url,
        json=flow_secrets,
        headers=headers(token),
    )
    resp.raise_for_status()


@with_token
def create_flow(token: str, name: str, workload: Dict[str, Any]) -> None:
    log.logger.debug(f"creating flow: {name}")
    url = f"{config.carrier_endpoint}/flow/{name}"
    if config.dc_region is not None:
        url += f"?region={config.dc_region}"
    hdrs = headers(token)
    log.logger.debug(f"creating flow\n{json.dumps(workload, indent=2)}")

    resp = requests.put(
        url,
        json=workload,
        headers=hdrs,
    )
    resp.raise_for_status()


@with_token
def delete_flow(token: str, name: str) -> None:
    log.debug(f"deleting flow: {name}")

    url = f"{config.carrier_endpoint}/flow/{name}"
    if config.dc_region is not None:
        url += f"?region={config.dc_region}"

    resp = requests.delete(
        url,
        headers=headers(token),
    )
    resp.raise_for_status()


# Note: These two functionts use KeyValueStorageAPI() for better exception handling.
def create_store(name: str, store_config: Optional[Dict[str, Any]] = None) -> None:
    """
    Creates a KV store.
    """
    log.logger.debug(f"creating store: {name}")
    if store_config is None:
        store_config = {}
    if config.region is not None:
        store_config["allow_locations"] = [f"region/{config.region}"]
    kv = KeyValueStorageAPI()
    kv.create_store(name, body=store_config)


def delete_store(name: str) -> None:
    """
    Deletes a KV store.
    """
    log.logger.debug(f"deleting store: {name}")
    kv = KeyValueStorageAPI()
    try:
        kv.delete_store(name)
    except ApiException as e:
        if e.status != 404:
            raise


def zip_current_directory(directory_name: str) -> str:
    current_directory = os.getcwd()
    zip_filename = os.path.join(current_directory, "build", "project_for_deploy.zip")

    with zipfile.ZipFile(zip_filename, "w", zipfile.ZIP_DEFLATED) as zipf:
        zipf.write("pyproject.toml", os.path.relpath("pyproject.toml", current_directory))

        env_file = os.environ.get("SEAPLANE_ENV_FILE", ".env")
        if os.path.exists(env_file):
            zipf.write(env_file, os.path.relpath(".env", current_directory))

        for root, _, files in os.walk(os.path.join(current_directory, directory_name)):
            for file in files:
                if "__pycache__" in root:
                    continue

                file_path = os.path.join(root, file)
                zipf.write(file_path, os.path.relpath(file_path, current_directory))

    return zip_filename


def upload_project(project_directory_name: str) -> str:
    """
    Zips the project directory and pushes it into the Seaplane object store,
    returning a URL that our executor image can use to refer back to the
    project when executing.

    project_directory_name should be the basename (no slashes) of the project
    directory, a peer to the pyproject.toml file for this Seaplane application.
    """
    # Step 1: Make sure we have a bucket to dump our project into
    default_bucket_name: str = "seaplane-internal-flows"
    default_bucket_config = {
        "description": "Seaplane bucket used for flow images. Should not be modified directly.",
        "replicas": 3,
        "max_bytes": -1,  # unlimited
        "allow_locations": ["all"],
    }

    obj = ObjectStorageAPI()
    obj._allow_internal = True
    if default_bucket_name not in obj.list_buckets():
        obj.create_bucket(default_bucket_name, default_bucket_config)

    # Step 2: Build the zip file
    project_file = zip_current_directory(project_directory_name)
    remote_path = project_directory_name + "." + _file_md5(project_file) + ".zip"

    # Step 3: Upload & return
    #  Retry upload if there is an exception (e.g., 500 timeout)
    for i in range(1, 4):
        try:
            obj.upload_file(default_bucket_name, remote_path, project_file)
            break
        except ApiException:
            time.sleep(i * 2)
            log.info(" retrying upload")

    obj_url = obj.file_url(default_bucket_name, remote_path)
    log.info(f"uploaded project package {obj_url}")

    return obj_url



import os
from typing import Any, Callable, Dict, List, Optional, Set, Union
from seaplane.config import config
from seaplane.errors import SeaplaneError

from seaplane.sdk_internal_utils.buckets import create_bucket_if_needed

from .executor import execute


"""
Deploy your applications to the Seaplane platform by using
Apps, Dags, and Flows.

# An App is a unit of deployment on Seaplane.
# Use an app to connect units of work together,
# and associate your application with Seaplane
# resources like HTTP endpoints and object store buckets.

app = seaplane.pipes.App("my-app")

# The work of apps is composed of Dags. Dags (Directed Acyclic Graphs)
# are collections of connections between the working python
# functions in your application. You can create an Dag using
# an application.

dag = app.dag("my-dag")

# You can use a dag to build Flows, which map to running
# processes on Seaplane infrastructure. To create a
# Flow, you need a dag, and some input sources, and a python
# function that does the work of the Flow.

def emphasize(context):
    yield context.body + b'!'

def intensify(context):
    yield context.body + 'b' for real'

# The first argument to dag.task is the function, the second is the input.

intensify_task = dag.task(intensify, [app.input()])
emphasize_task = dag.task(emphasize, [intensify_task])

# to associate an output with a dag, call `dag.respond()`

dag.respond(emphasize_task)

# to associate the output of a Flow or Dag with an App, do the same thing

app.respond(dag) # you can also write `app.respond(emphasize_task)` here

# Calling app.start() will deploy an application from your workstation,
# and *run* that application on the Seaplane platform.
app.start()
"""


class Subject:
    """
    Subject is a pubsub subject. Flows publish messages associated with
    a subject to communicate with the outside world.

    Subject should represent a concrete, publishable subject - it can
    contain fun ${! meta} macros, but should not contain wildcards.

    Flows can write to subjects but not subscribe to them.

    app.edge(my_task, Subject("known-stream.explicit-subject"))
    """

    def __init__(self, subject: str):
        self.subject = subject

    def __hash__(self) -> int:
        return hash(self.subject)

    def __eq__(self, other: Any) -> Any:
        return self.subject == other.subject


class Subscription:
    """
    The possibly-filtered name of a subject, for matching. Flows listen
    to subscriptions for their inputs.

    Should not ever contain macros and things, but may contain wildcards.

    Tasks can subscribe to Subscriptions but not write to them.

    app.edge(Subscription("known-stream.explicit-subscription.>"), my_task)
    """

    def __init__(self, filter: str, deliver: str = "all"):
        self.filter = filter
        self.deliver = deliver

    def __hash__(self) -> int:
        return hash(self.filter)

    def __eq__(self, other: Any) -> Any:
        return self.filter == other.filter

    def stream(self) -> str:
        return self.filter.split(".", 1)[0]


class _FlowSubject(Subject):
    """Flow subjects conform to a standard format

    app.task.address-tag.output_id.batch.batch.batch

    Because of this, they can be converted into subscriptions for peers.
    """

    def __init__(self, app_name: str, instance_name: str):
        self.app_name = app_name
        self.instance_name = instance_name
        super(_FlowSubject, self).__init__(
            f"{app_name}.{instance_name}"
            '.${! meta("_seaplane_address_tag") }'
            '.${! meta("_seaplane_output_id") }'
            '${! meta("_seaplane_batch_hierarchy") }'
        )

    def flow_subscription(self, address: str) -> Subscription:
        """use address = '*' to match all outputs from this task"""
        return Subscription(f"{self.app_name}.{self.instance_name}.{address}.>")


class Flow:
    """
    Description of deployment intent associated with a Seaplane flow.

    Create flows through an Dag, like this

    app = App("app")

    dag = app.dag("task-demo")

    def do_it(context: Messsage) -> Result:
        yield Result(b"Hello, " + context.body)

    my_task = dag.flow(do_it)

    Flow functions will be passed a Seaplane message, and
    should yield a Seaplane result. For convenience,
    if a task returns bytes (or something that can be converted
    into bytes), that value will be converted to bytes
    and packaged into a Result automatically.

    Once created, Flows can be wired together with other flows,
    application resources, or other Dags with `dag.edge`

    dag.edge(app.input(), my_task)
    dag.edge(my_task, app.output())
    """

    def __init__(
        self,
        dag: "Dag",
        work: Callable[..., Any],
        instance_name: str,
        subject: _FlowSubject,
        replicas: int,
        ack_wait_secs: int,
    ):
        # Note that app is a circular reference, so Tasks will not be GC'd
        self.dag = dag
        self.work = work
        self.instance_name = instance_name
        self.subject: Subject = subject  # task output
        self.replicas = replicas
        self.ack_wait_secs = ack_wait_secs
        self.subscriptions: Set[Subscription] = set()  # pull input from these subjects

    def subscribe(self, source: Subscription) -> None:
        self.subscriptions.add(source)


EdgeFrom = Union[Flow, "Dag", Subscription, "Bucket"]
EdgeTo = Union[Flow, Subject]


# For historical reasons, we need to be sure that
# these defaults work the same both for Dag.flow() and Dag.task()
_DEFAULT_REPLICAS = 1
_DEFAULT_ACK_WAIT_SECS = 2 * 60


class Dag:
    """
    Dag is a namespace and unit of deployment for
    a collection of Tasks. Dags create a namespace for task messages
    and manage deployment together.

    Create Dags through an application, like this

    app = App("my-app")
    dag = app.dag("my-dag")

    Dags are a namespace for their associated flows, so while the names
    of flows must be unique per dag, you can use the same flow name
    in different Dags. You can communicate between Dags either by
    wiring their flows together directly, or by calling `dag.respond()`
    with a flow that represents the single output of the dag (if that
    makes sense for your application)

    dag1 = app.dag("dag-1")
    dag2 = app.dag("dag-2")

    t1 = dag1.task(work1, [app.input()])
    dag1.respond(t1)

    # It's ok to connect tasks across Dag boundaries
    t2 = dag2.task(work2, [t1])

    # since we've configured dag1 to respond with t1's output, this works just the same
    # as the line above
    t2 = dag2.task(work2, [dag1])
    """

    def __init__(self, name: str):
        """use App.dag to create Dags"""
        self.name = name
        self.flow_registry: Dict[str, Flow] = {}
        self.response: Optional[Flow] = None

    def task(
        self,
        work: Callable[..., Any],
        sources: List[EdgeFrom],
        instance_name: Optional[str] = None,
        replicas: int = _DEFAULT_REPLICAS,
        ack_wait_secs: int = _DEFAULT_ACK_WAIT_SECS,
    ) -> Flow:
        """
        Dag.task() is a convenience method for describing a new flow, along
        with its inputs.

        Dag.task() returns a reference to the created flow, that can be used
        to subscribe subsequent flows to the created flow's output.
        """
        ret = self.flow(
            work, instance_name=instance_name, replicas=replicas, ack_wait_secs=ack_wait_secs
        )
        for source in sources:
            self.edge(source, ret)

        return ret

    def respond(self, tail: Union["Dag", Flow]) -> None:
        """
        Dag.respond() is a convenience for wiring tasks to
        the output of the dag.

        While you can wire flows together directly across dag boundaries,
        for many dags (that have a single, interesting output) it can
        be convenient to treat the dag itself as the source of messages.
        To do this, call `Dag.respond()` with the interesting output
        flow as an argument.

        Dags that have a response configured can be used as sources for
        other Flows.
        """
        if isinstance(tail, Dag):
            if tail.response is None:
                raise SeaplaneError(
                    f"{tail.name} has no response. Call Dag.respond on it, "
                    "or connect directly to it's tasks"
                )
            self.respond(tail.response)
        else:
            self.response = tail

    def flow(
        self,
        work: Callable[..., Any],
        instance_name: Optional[str] = None,
        replicas: int = _DEFAULT_REPLICAS,
        ack_wait_secs: int = _DEFAULT_ACK_WAIT_SECS,
    ) -> Flow:
        """
        Use Dag.flow to create an unwired flow, that can be associated
        with other flows by calling dag.edge().
        """
        if instance_name is None:
            instance_name = work.__name__.replace("_", "-") + "-default"

        prefixed_name = self.name + "-" + instance_name

        if prefixed_name in self.flow_registry:
            raise SeaplaneError(
                f"duplicate task name {prefixed_name} in dag {self.name}."
                " Try providing a unique instance_name argument when you create your task"
            )

        subject = _FlowSubject(self.name, prefixed_name)
        ret = Flow(self, work, prefixed_name, subject, replicas, ack_wait_secs)
        self.flow_registry[ret.instance_name] = ret
        return ret

    def edge(
        self,
        source: EdgeFrom,
        dest: EdgeTo,
    ) -> None:
        """
        Add an edge from a to b, binding b to a's output.

        Flows can be wired to publish information to other tasks and explicit subjects.
        Flows can be wired to recieve information from other tasks or subscriptions.

        If source is not a Flow or a Dag that responds with a task, then dest must be a Flow.
        """
        if isinstance(source, Dag):
            if source.response is None:
                raise SeaplaneError("Call Dag.response(..), or compose individual tasks together.")

            return self.edge(source.response, dest)

        if isinstance(source, Bucket):
            source = source.notify_subscription

        if not (isinstance(source, Flow) or isinstance(dest, Flow)):
            raise RuntimeError("edges must pass through a task")

        if isinstance(source, Subject):
            raise SeaplaneError(
                "the first argument to edge must be a Task, a Dag, or a Subscription"
            )

        if not (isinstance(dest, Flow) or isinstance(dest, Subject)):
            raise SeaplaneError("the second argument to edge must be a Task or a Subject")

        if isinstance(source, Flow) and isinstance(dest, Flow):
            if isinstance(source.subject, _FlowSubject):
                # TODO: Conditional dispatch will require addresses other than "*"
                source = source.subject.flow_subscription("*")
            else:
                raise SeaplaneError(
                    f"can't listen to task subject {source.subject.subject} automatically."
                    " To listen to this, you'll need to use an explicit Subscription"
                    " and write the Seaplane carrier filter yourself"
                )

        if isinstance(source, Flow) and isinstance(dest, Subject):
            source.subject = dest

        elif isinstance(source, Subscription) and isinstance(dest, Flow):
            dest.subscribe(source)
        else:
            # Don't trust the type system!
            raise RuntimeError(
                f"cannot wire {source} to {dest}."
                " The source of an edge must be a Subscription or a Flow"
            )


class Bucket:
    """
    A reference to a bucket and it's associated notification subscription.

    Get a bucket by asking the app to query the Seaplane infrastructure, like this:

    app = App("bucket-demo-app")
    dag = app.dag("dag")

    # This queries the Seaplane infrastructure, and may creaet a bucket
    # if one doesn't exist already
    bucket = app.bucket("bucket-demo-bucket")

    # once you have a bucket in hand, it can be used as the source
    # of events for a flow
    dag.task(observe_bucket, [bucket])
    """

    def __init__(self, name: str, notify: str):
        self.name = name
        self.notify_subscription = Subscription(notify)


class App:
    """
    An App is a unit of deployment to Seaplane. Use an App to create
    and deploy deploy Dags, which in turn you will use to create and
    organize individual tasks.

    An App is associated with a set of endpoints you can use to listen
    to HTTP requests and push data out of your Seaplane application, and
    optional references to Seaplane object store buckets.
    """

    class OutEndpoint(Subject):
        """
        The output seaplane endpoint. Does not accept subscriptions.

        Get this by calling App.output()
        """

        def __init__(self, endpoint: str):
            super(App.OutEndpoint, self).__init__(
                f"_SEAPLANE_ENDPOINT.out.{endpoint}."
                '${! meta("_seaplane_output_id") }'
                '${! meta("_seaplane_batch_hierarchy") }'
            )

    class InEndpoint(Subscription):
        """
        Endpoint represents the Subject of a seaplane input endpoint.

        Unlike most subscriptions, tasks subscribing to the seaplane input endpoint are
        configured to receive only new messages, rather than pull all messages
        from their upstream peers in order.

        Get one of these with App.input()
        """

        def __init__(self, endpoint: str):
            super(App.InEndpoint, self).__init__(
                f"_SEAPLANE_ENDPOINT.in.{endpoint}.>", deliver="new"
            )
            self.endpoint = endpoint

    # TODO: change the singleton to a REGISTRY approach, there is no real
    # TODO: need to enforce single-app processes.
    _instance: Optional["App"] = None

    def __init__(self, name: str, name_prefix: Optional[str] = None):
        if App._instance is not None:
            raise SeaplaneError(
                "you should only define a single app in your program."
                " {App._instance.name} is already defined"
            )

        if name_prefix is None:
            name_prefix = config.name_prefix

        self.name_prefix = name_prefix
        self.name = name_prefix + name
        self.dag_registry: Dict[str, Dag] = {}
        self.buckets: Set[Bucket] = set()
        self.input_endpoint = self.InEndpoint(self.name)
        self.output_endpoint = self.OutEndpoint(self.name)
        App._instance = self

    @classmethod
    def instance(cls) -> Optional["App"]:
        return cls._instance

    def bucket(self, bucket_name: str) -> Bucket:
        """
        Return a notify-capable bucket.

        Makes a network request to the Seaplane object store. Will
        attempt to create a new bucket if one doesn't exist, and may
        fail with an Exception if there is a service failure or if
        the named bucket exists but is not configured to send notifications.
        """

        notify = create_bucket_if_needed(self.name, bucket_name)
        b = Bucket(bucket_name, notify)
        self.buckets.add(b)
        return b

    def dag(self, dagname: str) -> Dag:
        prefixed_name = self.name_prefix + dagname
        if prefixed_name in self.dag_registry:
            raise SeaplaneError(
                f'there is already a dag named "{prefixed_name}" in application "{self.name}".'
                " The dags in an application must have unique names."
            )

        ret = Dag(prefixed_name)
        self.dag_registry[ret.name] = ret
        return ret

    def respond(self, tail: Union[Dag, Flow]) -> None:
        dag = tail.dag if isinstance(tail, Flow) else tail
        dag.edge(tail, self.output())

    def input(self) -> InEndpoint:
        """Return the "in" `Endpoint` for this Dag."""
        return self.input_endpoint

    def output(self) -> OutEndpoint:
        """Return the "out" `Endpoint` for this Dag."""
        return self.output_endpoint

    def run(self) -> None:
        """
        Run the application as the main work of a process.

        Reads command line arguments and the INSTANCE_NAME environment variable. If INSTANCE_NAME
        is present, and no arguments are provided, will attempt to execute the associated task.
        """
        # This has a bunch of circular dependencies but the ergonomics of `app.run()`
        # seem to make it worthwhile.
        import sys
        import toml

        import seaplane.deploy
        import seaplane.run_load_dotenv

        command = None
        if len(sys.argv) > 1:
            command = sys.argv[1]

        instance_name = os.getenv("INSTANCE_NAME")

        if command == "deploy":
            pyproject = toml.loads(open("pyproject.toml", "r").read())
            project_directory_name = pyproject["tool"]["poetry"]["name"]
            seaplane.deploy.deploy(self, project_directory_name)
        elif command == "destroy":
            seaplane.deploy.destroy(self)
        elif instance_name:
            task = None
            for dag in self.dag_registry.values():
                if instance_name in dag.flow_registry:
                    task = dag.flow_registry[instance_name]
                    break

            if task is None:
                raise RuntimeError(f"no task instance named {instance_name} found")

            execute(task.instance_name, task.work)



from importlib.metadata import version
import os
from typing import Dict, Optional

import seaplane_framework.api
import seaplane_framework.config

from seaplane.errors import SeaplaneError
from seaplane.logs import log
from seaplane.sdk_internal_utils.token_api import TokenAPI

import seaplane.run_load_dotenv  # noqa: F401

_SEAPLANE_IDENTIFY_API_ENDPOINT = "https://flightdeck.cplane.cloud/v1"
_SEAPLANE_GLOBAL_SQL_API_ENDPOINT = "https://sql.cplane.cloud/v1"
_SEAPLANE_SQL_DATABASE_ENDPOINT = "sql.cplane.cloud"
_SEAPLANE_CARRIER_API_ENDPOINT = "https://carrier.cplane.cloud/v1"
_SEAPLANE_SUBSTATION_EMBED_API_ENDPOINT = "https://embed.substation.cplane.cloud/v1/embed"
_SEAPLANE_VECTOR_DB_API_ENDPOINT = "https://vector-new.cplane.cloud"

api_key_names = ["SEAPLANE_API_KEY"]


# TODO get_api_keys and associated codepaths are unused, should be removed.
def get_api_keys() -> Dict[str, str]:
    api_keys = {}

    for api_key_env_name in api_key_names:
        api_key = os.environ.get(api_key_env_name)
        if api_key:
            api_keys[api_key_env_name] = api_key

    return api_keys


def runner_image() -> str:
    sdk_version = version("seaplane")
    return f"us-central1-docker.pkg.dev/artifacts-356722/sdk-apps/apps-executor:{sdk_version}"


class Configuration:
    """
    Seaplane SDK Configuration.

    Everytime the configuration is changed,
    It'll clear local configurations to the default Auth module.
    """

    def __init__(self) -> None:
        self.region: Optional[str] = None
        self.dc_region: Optional[str] = None
        self.identify_endpoint = _SEAPLANE_IDENTIFY_API_ENDPOINT
        self.global_sql_endpoint = _SEAPLANE_GLOBAL_SQL_API_ENDPOINT
        self.sql_database_endpoint = _SEAPLANE_SQL_DATABASE_ENDPOINT
        self.carrier_endpoint = _SEAPLANE_CARRIER_API_ENDPOINT
        self.vector_db_endpoint = _SEAPLANE_VECTOR_DB_API_ENDPOINT
        self.substation_embed_endpoint = _SEAPLANE_SUBSTATION_EMBED_API_ENDPOINT
        self.name_prefix = os.environ.get("SEAPLANE_NAME_PREFIX", "")
        self._api_keys: Dict[str, str] = get_api_keys()
        self._production = False

        seaplane_api_key = os.environ.get("SEAPLANE_API_KEY")
        self._token_api = TokenAPI(url=f"{self.identify_endpoint}/token", api_key=seaplane_api_key)

    def set_api_key(self, api_key: str) -> None:
        """Set the Seaplane API Key.

        The API Key is needed for the Seaplane Python SDK usage.

        Parameters
        ----------
        api_key : str
            Seaplane API Key.
        """
        self._token_api.set_api_key(api_key)

    def set_api_keys(self, api_keys: Dict[str, str]) -> None:
        """Set the Seaplane API Keys for Apps.

        The API Keys is needed for some of the Tasks.

        Supported Tasks API Keys:

        Seaplane: SEAPLANE_API_KEY

        For example, for use an OpenAI Task,
        you need to provide the Key - Value, of the API Key.

            $ from seaplane import sea

            $ api_keys = {"OPENAI_API_KEY": "sp-api-key-test" }
            $ sea.config.set_api_keys(api_keys)

        Parameters
        ----------
        api_keys : object
            API Keys and values.
        """

        if api_keys is None:
            raise SeaplaneError("api_keys parameters can't be None")
        self._api_keys = api_keys
        if api_keys.get("SEAPLANE_API_KEY", None) is not None:
            self.set_api_key(api_keys["SEAPLANE_API_KEY"])
            self._api_keys.pop("SEAPLANE_API_KEY")

    def set_identify_endpoint(self, endpoint: str) -> None:
        if endpoint[-1] == "/":
            self.identify_endpoint = endpoint.rstrip(endpoint[-1])
        else:
            self.identify_endpoint = endpoint

        self._token_api.set_url(url=f"{self.identify_endpoint}/token")

    def set_global_sql_endpoint(self, endpoint: str) -> None:
        if endpoint[-1] == "/":
            self.global_sql_endpoint = endpoint.rstrip(endpoint[-1])
        else:
            self.global_sql_endpoint = endpoint

    def set_sql_database_endpoint(self, endpoint: str) -> None:
        if endpoint[-1] == "/":
            self.sql_database_endpoint = endpoint.rstrip(endpoint[-1])
        else:
            self.sql_database_endpoint = endpoint

    def set_carrier_endpoint(self, endpoint: str) -> None:
        if endpoint[-1] == "/":
            self.carrier_endpoint = endpoint.rstrip(endpoint[-1])
        else:
            self.carrier_endpoint = endpoint

    def set_vector_endpoint(self, endpoint: str) -> None:
        if endpoint[-1] == "/":
            self.vector_db_endpoint = endpoint.rstrip(endpoint[-1])
        else:
            self.vector_db_endpoint = endpoint

    def set_substation_embed_endpoint(self, endpoint: str) -> None:
        if endpoint[-1] == "/":
            self.substation_embed_endpoint = endpoint.rstrip(endpoint[-1])
        else:
            self.substation_embed_endpoint = endpoint

    def log_level(self, level: int) -> None:
        """Change logging level.
        Seaplane uses Python logging module for internal logs.
        Python logging levels can be used directly with Seaplane Python SDK or
        use the already defined in seaplane.log module.
            $ from seaplane import sea, log
            $ sea.config.log_level(log.INFO)
        Parameters
        ----------
        level : int
            Logging Level from Python logging module,
            like DEBUG, INFO, WARNING, ERROR, CRITICAL
        """
        log.level(level)

        if level == log.DEBUG:
            log.debug("Seaplane debug activated")
            log.debug(f"Identify endpoint: {self.identify_endpoint}")
            log.debug(f"Carrier endpoint: {self.carrier_endpoint}")
            log.debug(f"Vector endpoint: {self.vector_db_endpoint}")

    def log_enable(self, enable: bool) -> None:
        """Enable or disable the Seaplane logging for the SDK.

        Parameters
        ----------
        enable : bool
            True to enable, False to disable.
        """
        if enable:
            log.enable()
        else:
            log.disable()

    def staging_mode(self) -> None:
        self.set_global_sql_endpoint("https://sql.staging.cplane.dev/v1")
        self.set_sql_database_endpoint("sql.staging.cplane.dev")
        self.set_carrier_endpoint("https://carrier.staging.cplane.dev/v1")
        self.set_identify_endpoint("https://flightdeck.staging.cplane.dev/v1")
        self.set_vector_endpoint("https://vector-new.staging.cplane.dev")

    def set_region(self, region: str) -> None:
        self.region = region.lower()
        if self.region in {"xa"}:
            self.dc_region = "sin"
        elif self.region in {"xe", "xf", "xu"}:
            self.dc_region = "fra"
        else:
            self.dc_region = "sjc"

    def get_platform_configuration(self) -> seaplane_framework.api.Configuration:
        configuration = seaplane_framework.api.Configuration()
        configuration.host = self.carrier_endpoint
        return configuration

    def get_access_token(self) -> str | None:
        return self._token_api.access_token


config = Configuration()



from importlib.metadata import version
from typing import Dict, Optional

SDK_HTTP_ERROR_CODE = 0


def headers(api_key: Optional[str], content_type: str = "application/json") -> Dict[str, str]:
    return {
        "Accept": "application/json",
        "Content-Type": content_type,
        "Authorization": f"Bearer {api_key}",
        "X-Seaplane-Sdk-Version": version("seaplane"),
    }



from seaplane.errors import SeaplaneError
from seaplane.kv import kv_store
from typing import Any, Dict, Generator

import json
import os
import requests

SUBSTATION_RESULTS_STREAM = "_SEAPLANE_AI_RESULTS"


class Substation:
    """
    Class for handling requests to and responses from Substation.

    See docstring for `make_request` for expected input and list of supported models.
    """

    def __init__(self, name: str):
        self.app_division = name
        self.request_store = f"_SP_REQUEST_{name}"
        self.response_store = f"_SP_RESPONSE_{name}"

    def results_stream(self) -> str:
        """
        Returns a string with the substation results stream
        """
        return SUBSTATION_RESULTS_STREAM

    def get_model(self, input_data: Dict[str, Any]) -> Dict[Any, Any]:
        """
        Returns headers and parameters for the selected model
        """
        model_name = input_data.get("model", "None").lower()
        headers = {"content-type": "application/json", "X-Division": self.app_division}
        data = {}

        # Default embeddings model (https://replicate.com/replicate/all-mpnet-base-v2)
        if model_name in (
            "replicate/all-mpnet-base-v2",
            "all-mpnet-base-v2",
            "embeddings",
        ):
            headers[
                "X-Version"
            ] = "b6b7585c9640cd7a9572c6e129c9549d79c9c31f0d3fdce7baac7c67ca38f305"
            for param in ("text", "text_batch"):
                value = input_data.get(param)
                if value:
                    data[param] = value

        # Zephyr-7B (https://replicate.com/tomasmcm/zephyr-7b-beta)
        elif model_name in (
            "tomasmcm/zephyr-7b-beta",
            "zephyr-7b-beta",
            "zephyr-7b",
        ):
            headers[
                "X-Version"
            ] = "961cd6665b811d0c43c0b9488b6dfa85ff5c7bfb875e93b4533e4c7f96c7c526"
            for param in (
                "prompt",
                "max_new_tokens",
                "temperature",
                "top_p",
                "top_k",
                "presence_penalty",
            ):
                value = input_data.get(param)
                if value:
                    data[param] = value

        # Mistral-7b-instruct (https://replicate.com/mistralai/mistral-7b-instruct-v0.1)
        elif model_name in (
            "mistralai/mistral-7b-instruct-v0.1",
            "mistral-7b-instruct-v0.1",
            "mistral-7b-instruct",
        ):
            headers[
                "X-Version"
            ] = "5fe0a3d7ac2852264a25279d1dfb798acbc4d49711d126646594e212cb821749"
            for param in (
                "prompt",
                "max_new_tokens",
                "min_new_tokens",
                "temperature",
                "top_p",
                "top_k",
                "repetition_penalty",
                "stop_sequences",
                "seed",
                "prompt_template",
            ):
                value = input_data.get(param)
                if value:
                    data[param] = value

        # BAAI/bge-large-en-v1.5 (https://replicate.com/nateraw/bge-large-en-v1.5)
        elif model_name in ("nateraw/bge-large-en-v1.5", "bge-large-en-v1.5"):
            headers[
                "X-Version"
            ] = "9cf9f015a9cb9c61d1a2610659cdac4a4ca222f2d3707a68517b18c198a9add1"
            for param in (
                "texts",
                "batch_size",
                "normalize_embeddings",
            ):  # leaving out path, convert_to_numpy as unsupported
                value = input_data.get(param)
                if value:
                    data[param] = value

        # CodeLlama-13b-Instruct (https://replicate.com/meta/codellama-13b-instruct)
        elif model_name in ("meta/codellama-13b-instruct", "codellama-13b-instruct"):
            headers[
                "X-Version"
            ] = "a5e2d67630195a09b96932f5fa541fe64069c97d40cd0b69cdd91919987d0e7f"
            for param in (
                "prompt",
                "system_prompt",
                "max_tokens",
                "temperature",
                "top_p",
                "top_k",
                "frequency_penalty",
                "presence_penalty",
                "repeat_penalty",
            ):
                value = input_data.get(param)
                if value:
                    data[param] = value

        # CodeLlama-34b-Python (https://replicate.com/meta/codellama-34b-python)
        elif model_name in ("meta/codellama-34b-python", "codellama-34b-python"):
            headers[
                "X-Version"
            ] = "e4cb69045bdb604862e80b5dd17ef39c9559ad3533e9fd3bd513cc68ff023656"
            for param in (
                "prompt",
                "max_tokens",
                "temperature",
                "top_p",
                "top_k",
                "frequency_penalty",
                "presence_penalty",
                "repeat_penalty",
            ):
                value = input_data.get(param)
                if value:
                    data[param] = value

        # Stable Diffusion 2.1 (https://replicate.com/stability-ai/stable-diffusion)
        elif model_name in (
            "stability-ai/stable-diffusion-2-1",
            "stability-ai/stable-diffusion",
            "stable-diffusion",
        ):
            headers[
                "X-Version"
            ] = "ac732df83cea7fff18b8472768c88ad041fa750ff7682a21affe81863cbe77e4"
            for param in (
                "prompt",
                "height",
                "width",
                "negative_prompt",
                "num_outputs",
                "num_inference_steps",
                "guidance_scale",
                "scheduler",
                "seed",
            ):
                value = input_data.get(param)
                if value:
                    data[param] = value

        # stability-ai/sdxl (https://replicate.com/stability-ai/sdxl)
        elif model_name in ("stability-ai/sdxl", "sdxl"):
            headers[
                "X-Version"
            ] = "39ed52f2a78e934b3ba6e2a89f5b1c712de7dfea535525255b1aa35c5565e08b"
            for param in (
                "prompt",
                "negative_prompt",
                "image",
                "mask",
                "width",
                "height",
                "num_outputs",
                "scheduler",
                "num_inference_steps",
                "guidance_scale",
                "prompt_strength",
                "seed",
                "refine",
                "high_noise_frac",
                "refine_steps",
                "apply_watermark",
                "lora_scale",
                "disable_safety_checker",
            ):
                value = input_data.get(param)
                if value:
                    data[param] = value

        # andreasjansson/stable-diffusion-inpainting
        #  (https://replicate.com/andreasjansson/stable-diffusion-inpainting)
        elif model_name in (
            "andreasjansson/stable-diffusion-inpainting",
            "stable-diffusion-inpainting",
        ):
            headers[
                "X-Version"
            ] = "e490d072a34a94a11e9711ed5a6ba621c3fab884eda1665d9d3a282d65a21180"
            for param in (
                "prompt",
                "negative_prompt",
                "image",
                "mask",
                "invert_mask",
                "num_outputs",
                "num_inference_steps",
                "guidance_scale",
                "seed",
            ):
                value = input_data.get(param)
                if value:
                    data[param] = value

        # openai/whisper (https://replicate.com/openai/whisper)
        elif model_name in ("openai/whisper", "whisper"):
            headers[
                "X-Version"
            ] = "4d50797290df275329f202e48c76360b3f22b08d28c196cbc54600319435f8d2"
            for param in (
                "audio",
                "transcription",
                "translate",
                "language",
                "temperature",
                "patience",
                "suppress_tokens",
                "initial_prompt",
                "condition_on_previous_text",
                "temperature_increment_on_fallback",
                "compression_ratio_threshold",
                "logprob_threshold",
                "no_speech_threshold",
            ):
                value = input_data.get(param)
                if value:
                    data[param] = value

        # ResNet-50 (https://replicate.com/replicate/resnet)
        elif model_name in ("replicate/resnet", "resnet", "resnet-50"):
            headers[
                "X-Version"
            ] = "dd782a3d531b61af491d1026434392e8afb40bfb53b8af35f727e80661489767"
            for param in ("image",):
                value = input_data.get(param)
                if value:
                    data[param] = value

        # DEFAULT: meta/llama-2-70b-chat (https://replicate.com/meta/llama-2-70b-chat)
        # elif model_name in ("meta/llama-2-70b-chat", "llama-2-70b-chat", "Llama-2 (70B)"):
        else:
            headers[
                "X-Version"
            ] = "02e509c789964a7ea8736978a43525956ef40397be9033abf9fd2badfe68c9e3"
            for param in (
                "prompt",
                "system_prompt",
                "max_new_tokens",
                "min_new_tokens",
                "temperature",
                "top_p",
                "top_k",
                "stop_sequences",
                "seed",
            ):
                value = input_data.get(param)
                if value:
                    data[param] = value

        return {"headers": headers, "data": data}

    def make_request(self, input_data: Dict[str, Any]) -> Dict[Any, Any]:
        """
        Makes the request to substation and returns the request information, including ID.

        `input_data` (Dict/JSON) must include `"model"` (see below) and at least one input:

          For LLMs, usually `"prompt"` and optional args, like `"temperature"`.

          For embeddings, `"text"` (string) or `"texts"`/`"text_batch"` (JSON list of strings).

        Supported models:
          `"all-mpnet-base-v2"` (also: `"embeddings"`) /
          `"zephyr-7b"` /
          `"mistral-7b-instruct"` /
          `"bge-large-en-v1.5"` /
          `"codellama-13b-instruct"` /
          `"codellama-34b-python"` /
          `"stable-diffusion-2-1"` /
          `"sdxl"` /
          `"stable-diffusion-inpainting"` /
          `"whisper"` /
          `"resnet-50"` /
          `"llama-2-70b-chat"` (default model)
        """
        model_params = self.get_model(input_data)

        proxy_addr = os.getenv("SEAPLANE_PROXY_ADDRESS", "localhost:4195")
        url = f"http://{proxy_addr}/predictions"

        resp = requests.post(url, headers=model_params["headers"], json=model_params["data"])
        if resp.status_code != 200:
            raise SeaplaneError("Error making substation request")

        request_data = {"request": resp.json()}
        return request_data

    def get_response(self, msg: Any) -> Generator[Any, None, None]:
        """
        Use this task to get the completed substation response
        """
        data = json.loads(msg.body)

        # If this is request data from the previous task it will have "request"
        #  See if there is matching response data in KV
        request = data.get("request")
        if request:
            if kv_store.exists(self.response_store, request["id"]):
                response = json.loads(kv_store.get_key(self.response_store, request["id"]))
                data.update({"response": response})
                kv_store.delete_key(self.response_store, request["id"])
                yield json.dumps(data).encode()
            else:
                # store request_id and batch_hierarchy for later output
                request.update(
                    {
                        "seaplane_request_id": msg.meta["_seaplane_request_id"],
                        "seaplane_batch_hierarchy": msg.meta["_seaplane_batch_hierarchy"],
                    }
                )
                kv_store.set_key(self.request_store, request["id"], json.dumps(request).encode())
                yield

        # If this is a response from Substation it will have "output"
        #  See if there matching request data in KV
        if "output" in data:
            if kv_store.exists(self.request_store, data["id"]):
                request = json.loads(kv_store.get_key(self.request_store, data["id"]))
                data.update({"request": request})
                kv_store.delete_key(self.request_store, data["id"])

                # restore the original request_id and batch_hierarchy
                seaplane_request_id = request.pop("seaplane_request_id")
                seaplane_batch_hierarchy = request.pop("seaplane_batch_hierarchy")

                ret = msg.result(json.dumps(data).encode())
                ret.output_id = seaplane_request_id
                ret.override_batch_hierarchy(seaplane_batch_hierarchy)
                yield ret
            else:
                kv_store.set_key(self.response_store, data["id"], json.dumps(data).encode())
                yield



from typing import Any, Dict, List

from seaplane.config import config
from seaplane.pipes import Dag


def build_debug_schema(dags: List[Dag]) -> Dict[str, Any]:
    """
    Constructs a JSON-friendly / simple type structure describing
    some parts of the application for use by associated tooling.
    """
    # This is a legacy debug schema, and it's structure shouldn't be
    # relied on.

    schema: Dict[str, Any] = {"apps": {}}

    for dag in dags:
        app_desc: Dict[str, Any] = {
            "id": dag.name,
            "tasks": [],
        }

        for task in dag.flow_registry.values():
            task_desc = {
                "id": task.instance_name,
                "name": task.work.__name__,
                "replicas": task.replicas,
            }

            app_desc["tasks"].append(task_desc)

        schema["apps"][dag.name] = app_desc

    schema["carrier_endpoint"] = config.carrier_endpoint
    schema["identity_endpoint"] = config.identify_endpoint

    return schema



import json
import os
import requests
from tabulate import tabulate
from typing import List

from seaplane.config import config
from seaplane.apps.cli import CliError
from seaplane.sdk_internal_utils.http import headers
from seaplane.sdk_internal_utils.token_auth import with_token

ENDPOINTS_STREAM = "_SEAPLANE_ENDPOINT"


@with_token
def get_messages(token: str, subject: str) -> str:
    """
    Returns a string with the number of messages in stream subject
      OR "Stream not found" if it does not exist
      OR "0" if the stream exists but subject does not
    """
    stream_name = subject.split(".")[0]
    url = f"{config.carrier_endpoint}/stream/{stream_name}"

    resp = requests.get(
        url,
        headers=headers(token),
    )
    if resp.status_code == 404:
        return "Stream not found"

    stream_info = json.loads(resp.content)
    if stream_info.get("details") is not None:
        if subject in stream_info["details"]["subjects"].keys():
            return str(stream_info["details"]["subjects"][subject])
    return "0"


@with_token
def get_status(token: str, flow_name: str) -> str:
    """
    Returns the status of a flow
      OR "N/A" if status not available
    """
    url = f"{config.carrier_endpoint}/flow/{flow_name}/status"
    if config.dc_region is not None:
        url += f"?region={config.dc_region}"
    resp = requests.get(
        url,
        headers=headers(token),
    )
    if resp.status_code == 200:
        status = json.loads(resp.content)
        replicas = 1
        dead = 0
        output: List[str] = []
        for alloc in status.keys():
            if status[alloc] != "dead":
                output.append(f"{replicas}: {status[alloc]}")
                # We could instead show the alloc id, but it gets busy...
                # output.append(f"{alloc}: {status[alloc]}")
                replicas += 1
            else:
                dead += 1
        if dead > 0:
            output.append(f"{dead} completed tasks")
        return "\n".join(output)
    return "N/A"


@with_token
def status(token: str) -> None:
    """
    Prints the status of resources associated with an app
    """
    try:
        schema_file = open(os.path.join("build", "schema.json"), "r")
        schema = json.load(schema_file)
    except Exception:
        raise CliError(
            "Cannot load build schema. Try moving to the directory where you deploy your app."
        )

    table_headers = ["Task Name", "Status", "Messages In", "Messages Out"]

    for app in schema["apps"]:
        table = []

        for task in schema["apps"][app]["tasks"]:
            flow_name = task["id"]

            flow_status = get_status(flow_name)

            url = f"{config.carrier_endpoint}/flow/{flow_name}"
            if config.dc_region is not None:
                url += f"?region={config.dc_region}"
            resp = requests.get(
                url,
                headers=headers(token),
            )
            flow_info = json.loads(resp.content)

            messages_in = "N/A"
            messages_out = "N/A"
            if "input" in flow_info.keys():
                input_subject = flow_info["input"]["carrier"]["subject"]
                messages_in = get_messages(input_subject)
                # Look at the second switched output for carrier
                output_subject = flow_info["output"]["switch"]["cases"][1]["output"]["carrier"][
                    "subject"
                ]
                messages_out = get_messages(output_subject)
            table.append([flow_name, flow_status, messages_in, messages_out])

        print("\n")
        # "fancy_grid" handles newlines better than "outline"
        print(tabulate(table, table_headers, tablefmt="fancy_grid"))

    print("NOTE: Message counts are currently unavailable for HTTP ENDPOINTS.")



import time
import inspect
import json
import types

from typing import Dict, Any, Callable, Iterable, Optional

from seaplane_framework.flow import processor
from seaplane.logs import log
from seaplane.object import object_store


class Result:
    """
    a Result is a package that allows work functions to return metadata along with
    their primary payload.

    Get one of these by calling Message.result(), like this:

    def with_fancy_result(message):
        ret = message.result("here's the body!")
        ret.address_tag("and-heres-the-address-tag")
        yield ret

    """

    def __init__(self, body: bytes, context: "Message"):
        self.body = body
        self.meta: Dict[str, str] = dict(context.meta)
        self.batch_hierarchy_override: Optional[str] = None

    @property
    def address_tag(self) -> Optional[str]:
        return self.meta.get("_seaplane_address_tag")

    @address_tag.setter
    def address_tag(self, tag: str) -> None:
        self.meta["_seaplane_address_tag"] = tag

    @address_tag.deleter
    def address_tag(self) -> None:
        del self.meta["_seaplane_address_tag"]

    @property
    def output_id(self) -> Optional[str]:
        return self.meta.get("_seaplane_output_id")

    @output_id.setter
    def output_id(self, val: str) -> None:
        self.meta["_seaplane_output_id"] = val

    @output_id.deleter
    def output_id(self) -> None:
        del self.meta["_seaplane_output_id"]

    def drop(self) -> None:
        self.meta["_seaplane_drop"] = "True"

    def override_batch_hierarchy(self, batch_hierarchy: str) -> None:
        """
        Disables automatic batch numbering, instead using
        exactly the user-provided string.

        Cannot be safely undone for a single response.
        """
        self.batch_hierarchy_override = batch_hierarchy
        del self.meta["_seaplane_batch_hierarchy"]

    @property
    def batch_hierarchy(self) -> str:
        if self.batch_hierarchy_override:
            return self.batch_hierarchy_override
        else:
            return self.meta["_seaplane_batch_hierarchy"]


class Message:
    """
    TaskContext is what a Task receives when running on the Seaplane platform.

    It contains methods to read the metadata associated with inbound messages,
    and read common messages types.
    """

    def __init__(self, body: bytes, meta: Dict[str, str]):
        self.body = body
        self.meta = types.MappingProxyType(meta)
        self._object_data: Optional[bytes] = None

    @property
    def request_id(self) -> str:
        return self.meta["_seaplane_request_id"]

    @property
    def object_data(self) -> bytes:
        """
        Attempts to interpret the body of the context
        as an object store event, and to syncronously load
        the associated object and return it.

        May throw exceptions if the message isn't an object store
        notification, if the object store is unavailable, or if
        the object store would otherwise throw an exception (for example,
        if the notification was for a deleted object.)
        """
        if self._object_data is not None:
            return self._object_data

        try:
            msg = json.loads(self.body)
            bucket = msg["Bucket"]
            obj = msg["Object"]
        except (json.JSONDecodeError, KeyError):
            log.logger.error(
                "it doesn't look like this message was from the object store."
                " Make sure your tasks are configured to listen to object store messages"
                " when using context.object_data"
            )
            raise

        self._object_data = object_store.download(bucket, obj)

        return self._object_data

    def result(self, body: Any) -> Result:
        """
        returns a result that can be used to set metadata for outgoing messages,
        including address tags for conditional dispatch.

        "body" will be coerced into bytes, and `result` will throw a TypeError
        if this is impossible.
        """
        if isinstance(body, str):
            body = bytes(body, encoding="utf-8")

        body_bytes = bytes(body)
        return Result(body_bytes, self)


def execute(log_name: str, work: Callable[..., Any]) -> None:
    """
    Execute the given function in an infinite loop, passing
    in new Seaplane messages and propegating results through the
    seaplane application. Never returns.

    Exceptions are logged but do not break the loop.
    """
    processor.start()
    log.logger.info(f"{log_name} ready for processing")

    while True:
        try:
            message = processor.read()

            if isinstance(message, bytes):
                log.logger.error(
                    "message isn't properly packaged."
                    f" Got: {message.decode('utf-8', errors='backslashescape')}"
                )
                continue

            log.logger.debug(f"processing {message.body}")

            if "_seaplane_output_id" not in message.meta:
                # This must be the first task in a smartpipe, so we have to get the
                # Endpoints API generated request ID from the incoming nats_subject.
                request_id = message.meta["nats_subject"].split(".")[
                    -1
                ]  # The Endpoints API always adds a request ID as the leaf

                # TODO what if this is an object update or some other weird business?
                # TODO (maybe object updates have something useful in their metadata?)
                message.meta["_seaplane_request_id"] = request_id
                message.meta["_seaplane_output_id"] = request_id

            if "_seaplane_address_tag" not in message.meta:
                message.meta["_seaplane_address_tag"] = "default"

            if "_seaplane_batch_hierarchy" not in message.meta:
                message.meta["_seaplane_batch_hierarchy"] = ""

            task_context = Message(message.body, message.meta.copy())
            result = work(task_context)

            # This complex return protocol is an attempt at a welcoming initial UX
            #
            # First, we decide if the user is yielding a list, returning a single value,
            # or declining to return
            #   If the user returns nothing or yields nothing, treat it like a drop.
            #   If the user returns a generator, iterate over it.
            #   If the user returns a non-None, non-generator, wrap it in a list and
            #      iterate over the singleton
            #
            # Now, DWIM the values that are yielded from the user's iterator into results
            #   If the value is a Result, use it
            #     Otherwise, try and coerce that value into bytes, and use that as the
            #     body of a Result
            #   If we have an explicit batch hierarchy override, use it, otherwise
            #     If _seaplane_batch_hierarchy exists at all in the result, append to it
            #       Otherwise ignore it.

            if inspect.isgenerator(result):
                gen: Iterable[Any] = result
            else:
                gen = [result]

            batch_id = 1
            for output in gen:
                if output is None:
                    output = task_context.result(b"None")
                    output.meta["_seaplane_drop"] = "True"

                if not isinstance(output, Result):
                    try:
                        output = task_context.result(output)
                    except TypeError:
                        log.logger.error(f"can't encode result {repr(output)} as bytes")
                        continue

                use_meta = output.meta.copy()

                if output.batch_hierarchy_override:
                    use_meta["_seaplane_batch_hierarchy"] = output.batch_hierarchy_override
                elif "_seaplane_batch_hierarchy" in use_meta:
                    use_meta["_seaplane_batch_hierarchy"] += f".{batch_id}"
                    batch_id += 1

                if use_meta.get("_seaplane_drop"):
                    log.logger.info(
                        f'dropping output for "{use_meta["_seaplane_output_id"]}"'
                        " at user request"
                    )
                else:
                    log.logger.debug(
                        f'served "{use_meta.get("_seaplane_output_id", "<UNKNOWN OUTPUT ID>")}'
                        f'{use_meta.get("_seaplane_batch_hierarchy", "<NO BATCH INFO>")}"'
                    )

                output_msg = processor._Msg(output.body, use_meta)
                processor.write(output_msg)

        except AssertionError:
            # Special case for ease in testing, and because we assume customer
            # code that fails assertions expects to crash.
            raise
        except Exception as e:
            log.logger.error(f"Error running Task {log_name}, sleeping", exc_info=e)

            # We should revisit this exception handling, but for now
            # we just try not to spam whatever went wrong.
            time.sleep(2)
        finally:
            log.logger.debug("flushing outputs")
            processor.flush()



from typing import cast, Any, Dict, List, Optional, TypedDict

import requests
from seaplane_framework.api.apis.tags import object_api
from seaplane_framework.api.model.bucket import Bucket

from seaplane.config import config
from seaplane.errors import HTTPError, SeaplaneError
from seaplane.sdk_internal_utils.http import headers
from seaplane.sdk_internal_utils.token_auth import get_pdk_client, method_with_token

SP_BUCKETS = ["seaplane-internal-flows"]


class ListObjectsMetadata(TypedDict):
    """
    Dictionary wrapping the metadata returned by the `list_objects()` endpoint.
    """

    name: str
    digest: str
    mod_time: int
    size: int


class ObjectStorageAPI:
    """
    Class for handling Object Storage API calls.
    """

    _allow_internal: bool
    """
    If set, allows the wrapper to manipulate Seaplane-internal buckets.

    Should not be set in customer code!
    """

    def __init__(self) -> None:
        self._allow_internal = False

    def get_object_api(self, access_token: str) -> object_api.ObjectApi:
        return object_api.ObjectApi(get_pdk_client(access_token))

    @method_with_token
    def list_buckets(self, token: str) -> List[str]:
        """
        Returns a list of buckets
        """
        api = self.get_object_api(token)
        list = []
        resp = api.list_buckets()
        for name, _ in sorted(resp.body.items()):
            if self._allow_internal or name not in SP_BUCKETS:
                list.append(name)

        return list

    @method_with_token
    def get_bucket(self, token: str, name: str) -> Dict[str, Any]:
        """
        Get metadata about a bucket
        """
        api = self.get_object_api(token)
        resp = api.get_bucket(path_params={"bucket_name": name})
        return {k: cast(str, v) for (k, v) in resp.body.items()}

    @method_with_token
    def create_bucket(self, token: str, name: str, body: Optional[Bucket] = None) -> None:
        """
        Create a new bucket

        Optional body argument can be used to configure the bucket with description, notify, etc.
        """
        if not self._allow_internal and name in SP_BUCKETS:
            raise SeaplaneError(f"Cannot create bucket with Seaplane-internal name `{name}`")

        if not body:
            body = {}

        api = self.get_object_api(token)
        path_params = {
            "bucket_name": name,
        }
        api.create_bucket(
            path_params=path_params,
            body=body,
        )

    @method_with_token
    def delete_bucket(self, token: str, name: str) -> None:
        """
        Delete a bucket
        """
        if not self._allow_internal and name in SP_BUCKETS:
            raise SeaplaneError(f"Cannot delete bucket with Seaplane-internal name `{name}`")
        api = self.get_object_api(token)
        path_params = {
            "bucket_name": name,
        }
        api.delete_bucket(path_params=path_params)

    def exists(self, bucket: str, object: str) -> bool:
        """
        Returns True if object exists in bucket
        """
        try:
            objs = self.list_objects(bucket, "")
            for obj in objs:
                if obj["name"] == object:
                    return True
            return False
        except HTTPError as e:
            if e.status == 404:
                return False
            raise e

    @method_with_token
    def list_objects(
        self, token: str, bucket_name: str, path_prefix: str = "/"
    ) -> List[ListObjectsMetadata]:
        """
        List objects in a bucket, (optional) matching a path prefix

        Returns a list of dicts with name, digest, mod_time, and bytes for each object
        """
        if not self._allow_internal and bucket_name in SP_BUCKETS:
            raise SeaplaneError(
                f"Cannot list objects in bucket with Seaplane-internal name `{bucket_name}`"
            )
        api = self.get_object_api(token)

        path_params = {
            "bucket_name": bucket_name,
        }
        query_params = {
            "path": path_prefix,
        }
        resp = api.list_objects(
            path_params=path_params,
            query_params=query_params,
        )

        table = [
            ListObjectsMetadata(
                name=x["name"],
                digest=x["digest"],
                mod_time=x["mod_time"],
                size=x["size"],
            )
            for x in resp.body
        ]

        return table

    @method_with_token
    def download(self, token: str, bucket_name: str, path: str) -> bytes:
        """
        Downloads an object

        Returns the object in bytes
        """
        url = f"{config.carrier_endpoint}/object/{bucket_name}/store"

        params: Dict[str, Any] = {}
        params["path"] = path
        resp = requests.get(
            url,
            params=params,
            headers=headers(token, "application/octet-stream"),
        )
        return resp.content

    def file_url(self, bucket_name: str, path: str) -> str:
        """
        Builds a URL usable to download the object stored at the given bucket & path.
        """
        return f"{config.carrier_endpoint}/object/{bucket_name}/store?path={path}"

    @method_with_token
    def upload(self, token: str, bucket_name: str, path: str, object: bytes) -> None:
        """
        Create a new object from the given object data (bytes)
        """
        if not self._allow_internal and bucket_name in SP_BUCKETS:
            raise SeaplaneError(
                f"Cannot upload to bucket with Seaplane-internal name `{bucket_name}`"
            )
        api = self.get_object_api(token)

        path_params = {
            "bucket_name": bucket_name,
        }
        query_params = {
            "path": path,
        }

        api.create_object(
            path_params=path_params,
            query_params=query_params,
            body=object,
        )

    def upload_file(self, bucket_name: str, path: str, object_path: str) -> None:
        """
        Upload a local file to a new object
        """
        if not self._allow_internal and bucket_name in SP_BUCKETS:
            raise SeaplaneError(
                f"Cannot upload to bucket with Seaplane-internal name `{bucket_name}`"
            )
        with open(object_path, "rb") as file:
            file_data = file.read()

        self.upload(bucket_name, path, file_data)

    @method_with_token
    def delete(self, token: str, bucket_name: str, path: str) -> None:
        """
        Delete an object
        """
        if not self._allow_internal and bucket_name in SP_BUCKETS:
            raise SeaplaneError(
                f"Cannot delete from bucket with Seaplane-internal name `{bucket_name}`"
            )
        api = self.get_object_api(token)
        path_params = {
            "bucket_name": bucket_name,
        }
        query_params = {
            "path": path,
        }
        api.delete_object(
            path_params=path_params,
            query_params=query_params,
        )


object_store = ObjectStorageAPI()



import functools
from typing import Any, Callable, Generator, Iterable, List, Optional, Protocol, cast
from warnings import warn
from contextlib import contextmanager

from seaplane.logs import log
from seaplane.pipes import App, Flow, EdgeFrom


class LegacyAppsCompat:
    """
    a LEGACY application. New applications should use the new app toolkit
    """

    def __init__(
        self,
        func: Callable[[Any], Any],  # user code that assembles the task DAG.
        type: str,  # "stream" or "API"
        parameters: List[str],  # when type == "stream", [STREAM_NAME]
        id: str,
    ) -> None:
        self.id = id
        self.func = func
        self.real_app = App(self.id)
        self.real_dag = self.real_app.dag(self.id)

        if type != "API" or len(parameters) > 0:
            raise RuntimeError(
                "'stream' type apps are no longer supported."
                " To listen to arbitrary streams, use the new app construction tools"
            )

    def __call__(*args: Any, **kwargs: Any) -> None:
        raise RuntimeError(
            "Don't call your legacy app directly. Instead, call seaplane.apps.start()"
        )

    def assemble(self) -> None:
        """
        Executes the app function, which in turn calls task.wire_up to assemble
        task sources.
        """
        with legacy_context.using_current_app(self):
            last = self.func(self.real_app.input())

        if last is not None:
            self.real_dag.edge(last, self.real_app.output())


class LegacyTaskCompat:
    """
    Tasks are weird use / mention hybrids: in a deployment
    context, they are a map of Seaplane carrier flows and
    the connections between them. In an execution context
    on the Seaplane platform, they're containers for executable
    code that get the results of endpoints or tasks as input and can
    emit good things to their downstream flows and endpoints.
    """

    def __init__(
        self,
        func: Callable[[Any], Any],
        id: str,
        ack_wait: int = 8,
        replicas: int = 1,
        watch_bucket: Optional[str] = None,  # DO NOT USE
    ) -> None:
        self.func = func
        self.id = id
        self.name = func.__name__
        self.replicas = replicas
        self.ack_wait_secs = ack_wait * 60

        if watch_bucket is not None:
            raise RuntimeError("to use a watch_bucket, use the new application construction style")

    def wire_up(self, sources: Iterable[EdgeFrom]) -> "Flow":
        """
        Used when we're constructing a dag out of the tasks.
        """
        assert legacy_context.current_legacy_app is not None

        dag = legacy_context.current_legacy_app.real_dag

        instance = dag.flow(
            self.func,
            instance_name=self.id,
            replicas=self.replicas,
            ack_wait_secs=self.ack_wait_secs,
        )
        for src in sources:
            dag.edge(src, instance)

        return instance


class LegacyContext:
    """
    Built up from the various legacy @app and @task decorators in the code.

    New code should not rely on this, instead use a seaplane.pipes.Dag object to
    provide context for individual deployments and executions.
    """

    def __init__(self) -> None:
        self.apps: List[LegacyAppsCompat] = []
        self.tasks: List[LegacyTaskCompat] = []
        self.current_legacy_app: Optional[LegacyAppsCompat] = None

    def add_app(self, app: LegacyAppsCompat) -> None:
        self.apps.append(app)
        log.debug(f"App: {app.id}")

    @contextmanager
    def using_current_app(self, app: LegacyAppsCompat) -> Generator[None, None, None]:
        self.current_legacy_app = app
        yield
        self.current_legacy_app = None


legacy_context = LegacyContext()


def app(
    type: str = "API",
    parameters: Optional[List[str]] = None,
    id: Optional[str] = None,
    path: Optional[str] = None,
    _func: Optional[Callable[[Any], Any]] = None,
) -> Callable[[Callable[..., Any]], LegacyAppsCompat]:
    if path is not None:
        warn("The app path argument is deprecated", DeprecationWarning, stacklevel=2)
    if parameters is None:
        parameters = []

    if type != "API" or parameters != []:
        raise RuntimeError(
            "non API apps are not supported."
            " To listen to arbitrary streams, please use the new app construction tools."
        )

    def wrap_legacy_app_function(func: Callable[[Any], Any]) -> LegacyAppsCompat:
        set_id = id if id is not None else func.__name__.replace("_", "-")
        legacy_app = LegacyAppsCompat(func=func, id=set_id, type=type, parameters=parameters)
        legacy_context.add_app(legacy_app)

        return legacy_app

    if not _func:
        return wrap_legacy_app_function
    else:
        return wrap_legacy_app_function(_func)  # type: ignore


class TakeEdgeFroms(Protocol):
    def __call__(self, *args: EdgeFrom) -> Flow:
        ...


def task(
    id: Optional[str] = None,
    watch_bucket: Optional[str] = None,
    type: Optional[str] = None,  # UNUSED
    model: Optional[str] = None,  # UNUSED
    index_name: Optional[str] = None,  # UNUSED
    replicas: int = 1,
    ack_wait: int = 2,
    _func: Optional[Callable[[Any], Any]] = None,
) -> Callable[[Callable[..., Any]], TakeEdgeFroms]:
    deprecated_args = {"type": type, "model": model, "index_name": index_name}
    for arg in deprecated_args:
        if deprecated_args[arg] is not None:
            warn(f"The task {arg} argument is deprecated", DeprecationWarning, stacklevel=2)

    def decorator_task(func: Callable[..., Any]) -> TakeEdgeFroms:
        task_id = id if id is not None else func.__name__.replace("_", "-")

        task = LegacyTaskCompat(
            func=func,
            id=task_id,
            replicas=replicas,
            ack_wait=ack_wait,
            watch_bucket=watch_bucket,
        )

        @functools.wraps(func)
        def wrapper(*sources: EdgeFrom) -> Flow:
            return task.wire_up(sources)

        return cast(TakeEdgeFroms, wrapper)

    if not _func:
        return decorator_task
    else:
        return decorator_task(_func)  # type: ignore



from typing import List

from langchain.embeddings.base import Embeddings
from langchain.vectorstores import Qdrant
from pydantic import BaseModel
import requests

from seaplane.config import config
from seaplane.sdk_internal_utils.http import headers
from seaplane.sdk_internal_utils.token_auth import method_with_token
from seaplane.vector import vector_store


class SeaplaneEmbeddingFunction(BaseModel, Embeddings):
    @method_with_token
    def _embed(self, token: str, query: str) -> List[float]:
        url = config.substation_embed_endpoint

        resp = requests.post(
            url,
            headers=headers(token),
            json={"query": query},
        )
        resp.raise_for_status()
        result = resp.json()

        return [float(i) for i in result["embeddings"]]

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        embeddings = []
        for text in texts:
            response = self._embed(f"Represent the document for Retrieval: {text}")
            embeddings.append(response)
        return embeddings

    def embed_query(self, query: str) -> List[float]:
        response = self._embed(f"Represent the Science sentence: {query}")
        return response


seaplane_embeddings = SeaplaneEmbeddingFunction()


def langchain_vectorstore(index_name: str, embeddings: Embeddings = seaplane_embeddings) -> Qdrant:
    vectorstore = Qdrant(
        client=vector_store._get_client(),
        collection_name=index_name,
        embeddings=embeddings,
    )

    return vectorstore



from seaplane.kv import KeyValueStorageAPI
from seaplane.pipes.executor import Result

import json
import re
import time
from typing import Any, Generator, List

COLLECT_STORE = "_SP_COLLECT_"

# TODO: add option to compress messages (on stream AND kv store)


class Collector:
    """
    Class for handling debatching and collecting
    """

    def __init__(self, name: str):
        self.collect_store = f"{COLLECT_STORE}{name}"

    def debatch(self, msg: Any, input_list: List[Any]) -> List[Result]:
        """
        Splits a list into separate messages to be yielded in your task

        msg is the input to the task (to preserve metadata)

        input_list is a python List (e.g., of strings or JSON objects)

        Example:

        for ret in collector.debatch(msg, input_data["input"]):
            yield ret
        """
        total = len(input_list)
        print(f"processing list with {total} items...")
        msg_list: List[Result] = []
        for count in range(total):
            # this should cover anything in a JSON array deserialized with json.loads()
            if type(input_list[count]) is str:
                ret_bytes = input_list[count].encode()
            elif type(input_list[count]) is dict:
                ret_bytes = json.dumps(input_list[count]).encode()
            ret = msg.result(ret_bytes)
            ret.meta["_seaplane_batch_total"] = total
            ret.meta["_seaplane_debatch_hierarchy"] = msg.meta["_seaplane_batch_hierarchy"]
            msg_list.append(ret)
        return msg_list

    def collect(self, msg: Any) -> Generator[Any, None, None]:
        """
        Use this task to collect the responses and yield a JSON-compatible list of strings.

        Note: if your debatched data is binary (images, pickles, etc.) you can store them in
        object or KV store and yield the data (url or object name) necessary for retrieval.
        This is also preferred if the data is very large, to avoid message size limits.
        """

        # lots of stuff to preserve message hierarchy for later output
        request_id = msg.meta["_seaplane_request_id"]
        batch_total = int(msg.meta.get("_seaplane_batch_total", "0"))
        batch_hierarchy = msg.meta["_seaplane_batch_hierarchy"]
        debatch_hierarchy = msg.meta.get("_seaplane_debatch_hierarchy", "")
        message_id = f"{request_id}{batch_hierarchy.replace('.','_')}"

        print(f"received {message_id}/{batch_total}")

        kv = KeyValueStorageAPI()
        kv.set_key(self.collect_store, message_id, msg.body)
        time.sleep(1)  # should have consistency, but let's avoid getting throttled
        stored_keys = [
            x
            for x in kv.list_keys(self.collect_store)
            if re.match(f"{request_id}{debatch_hierarchy.replace('.','_')}.*", x)
        ]

        # once we have all of the keys stored we can put it all together
        if len(stored_keys) == batch_total:
            print(f"collecting request_id {request_id}...")
            collected = []
            debatch_list = batch_hierarchy.split(".")[len(debatch_hierarchy.split(".")) :]
            for i in range(1, batch_total + 1):
                debatch_list[0] = str(i)
                key = f"{request_id}{debatch_hierarchy.replace('.','_')}_{'_'.join(debatch_list)}"
                collected.append(kv.get_key(self.collect_store, key).decode())
            print(collected)
            ret = msg.result(json.dumps(collected))
            debatch_list[0] = str(1)
            new_batch_hierarchy = f"{debatch_hierarchy}.{'.'.join(debatch_list)}"
            ret.override_batch_hierarchy(new_batch_hierarchy)
            yield ret
        else:
            yield



from seaplane.config import config as global_config, Configuration

from seaplane.sdk_internal_utils.token_auth import method_with_token


class SQLAPI:
    """
    Class for accessing the SQL API
    """

    def __init__(self, config: Configuration) -> None:
        self.config = config

    def _connection_string(self, database_name: str) -> str:
        return f"postgres://{database_name}:{self.config.get_access_token()}@{self.config.sql_database_endpoint}/{database_name}"  # noqa: E501

    @method_with_token
    def connection_string(self, token: str, database_name: str) -> str:
        """
        Get a Seaplane SQL connection string for a given database
        """
        return self._connection_string(database_name)


sql = SQLAPI(global_config)



import os
import shutil
from typing import Any, Dict, List
from urllib.parse import urlparse

from seaplane.logs import log
from seaplane.pipes import App, Flow
from seaplane.config import config, runner_image

from .utils import (
    add_secrets,
    create_flow,
    create_stream,
    delete_flow,
    delete_stream,
    create_store,
    delete_store,
    upload_project,
)

"""
Tools for taking a fully constructed Dag / Task complex and deploying it
into the Seaplane infrastructure.

Call `deploy(app, project_directory_name)` to push your application to Seaplane.
"""


SP_UTIL_STORES = ["_SP_REQUEST_", "_SP_RESPONSE_", "_SP_COLLECT_"]


class ProcessorInfo:
    """Information about the docker container to run applications"""

    def __init__(self, runner_image: str, runner_args: List[str]):
        self.runner_image = runner_image
        self.runner_args = runner_args


def flow_into_config(flow: Flow, processor_info: ProcessorInfo) -> Dict[str, Any]:
    """Produces JSON.dump-able flow configuration suitable for running the given task"""

    ret = {
        "processor": {
            "docker": {
                "image": processor_info.runner_image,
                "args": processor_info.runner_args,
            }
        },
        "output": {
            "switch": {
                "cases": [
                    {
                        "check": 'meta("_seaplane_drop") == "True"',
                        "output": {"drop": {}},
                    },
                    {
                        "check": 'meta("_seaplane_drop") != "True"',
                        "output": {
                            "carrier": {
                                "subject": flow.subject.subject,
                            }
                        },
                    },
                ]
            }
        },
        "replicas": flow.replicas,
    }

    if len(flow.subscriptions) == 0:
        log.logger.warning(
            f"task {flow.instance_name} does not appear to consume any input, it may not run"
        )

    broker: List[Dict[str, Any]] = []
    for ix, src in enumerate(sorted(flow.subscriptions, key=lambda s: s.filter)):
        # this durable scheme means we're committed to destroying
        # the consumers associated with these flows on redeploy.
        # Future fancy hot-swap / live-update schemes may need
        # to use different durable names
        durable = f"{flow.instance_name}-{ix}"
        broker.append(
            {
                "carrier": {
                    "ack_wait": f"{flow.ack_wait_secs}s",
                    "bind": True,
                    "deliver": src.deliver,
                    "durable": durable,
                    "stream": src.stream(),
                    "subject": src.filter,
                },
            }
        )

    ret["input"] = {"broker": broker}

    return ret


def deploy(app: App, project_directory_name: str) -> None:
    """
    Runs a complete deploy of a given app.

    project_directory_name is the name of the directory, a peer to the
    pyproject.toml, that contains

    pyproject = toml.loads(open("pyproject.toml", "r").read())
    project_directory_name = pyproject["tool"]["poetry"]["name"]

    Will delete and recreate a "build" directory

    """
    shutil.rmtree("build/", ignore_errors=True)
    os.makedirs("build")

    project_url = upload_project(project_directory_name)
    processor_info = ProcessorInfo(runner_image(), [project_url])

    for bucket in app.buckets:
        delete_stream(bucket.notify_subscription.stream())
        create_stream(bucket.notify_subscription.stream())

    for base_store_name in SP_UTIL_STORES:
        store_name = base_store_name + app.name
        delete_store(store_name)
        create_store(store_name, store_config={"replicas": 1})

    for dag in app.dag_registry.values():
        delete_stream(dag.name)
        create_stream(dag.name)

        for task in dag.flow_registry.values():
            new_flow_config = flow_into_config(task, processor_info)
            delete_flow(task.instance_name)
            create_flow(task.instance_name, new_flow_config)

            secrets = {"INSTANCE_NAME": task.instance_name} | config._api_keys
            add_secrets(task.instance_name, secrets)

    log.logger.info(
        "to write to the application endpoint,\n"
        f"post to https://{urlparse(config.carrier_endpoint).netloc}"
        f"/v1/endpoints/{app.input().endpoint}/request\n"
        "or run: seaplane request -d <data>"
    )


def destroy(app: App) -> None:
    for dag in app.dag_registry.values():
        delete_stream(dag.name)
        for task in dag.flow_registry.values():
            delete_flow(task.instance_name)

    for base_store_name in SP_UTIL_STORES:
        store_name = base_store_name + app.name
        delete_store(store_name)



from typing import Dict, Optional, cast

import requests

from seaplane.errors import HTTPError
from seaplane.logs import log
from seaplane.sdk_internal_utils.http import SDK_HTTP_ERROR_CODE, headers


class TokenAPI:
    """
    Manage access token.

    Seaplane Python SDK manages the token by default,
    It can be managed manually as well.

    Any configuration change to the default `config` module,
    It'll reset `TokenAPI` local configurations, and renewed tokens.
    """

    def __init__(
        self,
        url: str,
        api_key: Optional[str],
        access_token: Optional[str] = None,
        auto_renew: bool = True,
        tenant: Optional[str] = None,
    ) -> None:
        self.url = url
        self.api_key = api_key
        self.access_token = access_token
        self.auto_renew = auto_renew
        self.tenant = tenant

    def __repr__(self) -> str:
        return f"TokenAPI(api_key={self.api_key}, access_token={self.access_token})"

    def set_url(self, url: str) -> None:
        if self.url == url:
            return
        self.url = url
        self.access_token = None
        self.tenant = None

    def set_api_key(self, key: str) -> None:
        if self.api_key == key:
            return
        self.api_key = key
        self.access_token = None
        self.tenant = None

    def renew_token(self) -> str:
        """Renew the token.

        Any configuration change to the default `config` module,
        It'll the renewed token, and It's needed to renew the token again.

        Returns
        -------
        str
            Returns the renewed token.
        """
        self.access_token = None

        payload = self._request_access_token()
        self.access_token = payload["token"]
        self.tenant = payload["tenant"]
        return self.access_token

    def get_token(self) -> str:
        """
        Returns the current access token, requesting a new one if needed.
        """
        if self.access_token is None:
            return self.renew_token()
        return self.access_token

    def get_tenant(self) -> str:
        """Request the current tenant

        Returns
        -------
        str
            Returns the tenant.
        """
        if not self.tenant:
            self.renew_token()
        # A `self.renew_token()` call will always set the tenant variable from
        # the newly-renewed token, so we can always return it here.
        return self.tenant or ""

    def _request_access_token(self) -> Dict[str, str]:
        try:
            log.info("Requesting access token...")
            if not self.api_key:
                log.error("API KEY not set, use seaplane.config.set_api_key")

            response = requests.post(self.url, json={}, headers=headers(self.api_key))

            if response.ok:
                payload = cast(Dict[str, str], response.json())
                return payload
            else:
                error_body = response.text
                log.error(
                    f"Bad Access token request code {response.status_code}, error {error_body}"
                )
                raise HTTPError(response.status_code, error_body)

        except requests.exceptions.RequestException as err:
            log.error(f"Request access token exception: {str(err)}")
            raise HTTPError(SDK_HTTP_ERROR_CODE, str(err))



"""
command line interface
"""
from argparse import ArgumentParser
from importlib.metadata import version
import json
import os
import re
import requests
import toml
from typing import Any, Dict, cast

from cookiecutter.main import cookiecutter

import seaplane.run_load_dotenv as denv
from seaplane.sdk_internal_utils.http import headers


PROJECT_TOML = "pyproject.toml"


class CliError(Exception):
    pass


def get_project() -> Dict[str, Any]:
    """
    throws an error if the project directory isn't structured in ways
    we expect it to be.
    """
    if not os.path.exists(PROJECT_TOML):
        raise CliError(
            f"{PROJECT_TOML} file missing, run seaplane init"
            " (or change directory to the root of your Seaplane project)"
        )

    project = toml.loads(open(PROJECT_TOML, "r").read())
    project_name = project["tool"]["poetry"]["name"]
    main = project["tool"]["seaplane"].get("main", None)

    if not os.path.exists(project_name):
        raise CliError(
            f"source file {project_name} directory missing, \
                the source code has to live under {project_name} directory."
        )

    if not project_name or not main:
        raise CliError(f"{PROJECT_TOML} not a valid Seaplane project.")

    return project


def cli_deploy() -> None:
    """
    Deploys an app
    """
    project = get_project()
    project_name = project["tool"]["poetry"]["name"]
    main_script = project["tool"]["seaplane"]["main"]
    os.system(f"poetry run python {project_name}/{main_script} deploy")


def cli_destroy() -> None:
    """
    Destroys an app (deletes streams and flows)
    """
    project = get_project()
    project_name = project["tool"]["poetry"]["name"]
    main_script = project["tool"]["seaplane"]["main"]
    os.system(f"poetry run python {project_name}/{main_script} destroy")


def cli_build() -> None:
    """
    Builds an app (writes build/schema.json)
    """
    project = get_project()
    project_name = project["tool"]["poetry"]["name"]
    main_script = project["tool"]["seaplane"]["main"]
    os.system(f"poetry run python {project_name}/{main_script} build")


def cli_status(watch: bool) -> None:
    """
    Provides status of an app's resources.
      If watch is enabled it must be in your path.
    """
    project = get_project()
    project_name = project["tool"]["poetry"]["name"]
    main_script = project["tool"]["seaplane"]["main"]
    if watch:
        os.system(f"watch poetry run python {project_name}/{main_script} status")
    else:
        os.system(f"poetry run python {project_name}/{main_script} status")


def get_token(fd_url: str) -> str:
    """
    Gets a token from Flightdeck
    """
    # Try to get the API key from .env
    api_key = os.environ.get("SEAPLANE_API_KEY", "None")
    if denv.loaded is False or api_key == "None":
        raise CliError("Set your API key in .env")

    response = requests.post(fd_url, json={}, headers=headers(api_key))
    if response.ok:
        payload = cast(Dict[str, str], response.json())
    else:
        error_body = response.text
        raise CliError(f"Bad Access token request code {response.status_code}, error {error_body}")

    return payload["token"]


def cli_request(data: str) -> None:
    """
    Performs an endpoint request, writes and prints the request_id
    """
    if data[0] == "@":
        try:
            data_bytes = open(data[1:], "rb").read()
        except Exception:
            raise CliError("Cannot load data file.")
    else:
        data_bytes = data.encode()

    # Load schema file
    try:
        schema_file = open(os.path.join("build", "schema.json"), "r")
        schema = json.load(schema_file)
    except Exception:
        raise CliError(
            "Cannot load build schema. Try moving to the directory where you deploy your app."
        )

    # Get a token
    fd_url = schema["identity_endpoint"] + "/token"
    token = get_token(fd_url)

    # Make the request
    for app in schema["apps"]:
        app_id = schema["apps"][app]["id"]
    carrier_endpoint = schema["carrier_endpoint"]
    url = f"{carrier_endpoint}/endpoints/{app_id}/request"
    bearer = {"Authorization": f"Bearer {token}", "Content-Type": "application/octet-stream"}
    resp = requests.post(url, headers=bearer, data=data_bytes)

    # Empty response handling
    if not resp.content:
        raise CliError(f"Error: empty response from endpoint (code={resp.status_code})")

    # Invalid response format handling
    try:
        resp_content = json.loads(resp.content)
    except json.decoder.JSONDecodeError:
        raise CliError(
            f"Error: invalid JSON response from endpoint (code={resp.status_code}): {resp.content!r}"  # noqa: E501
        )

    # Unexpected response code handling
    if resp.status_code < 200 or resp.status_code >= 300:
        raise CliError(
            f"Error: unexpected response code {resp.status_code} from endpoint: {resp.content!r}"
        )

    # Malformed response handling
    if "request_id" not in resp_content:
        raise CliError(
            f"Error: invalid response from endpoint, missing request_id (code={resp.status_code})"
        )

    request_id = resp_content["request_id"]

    # save the request_id to a file
    with open(".request_id", "a") as f:
        f.write(f"{request_id}\n")
    print(f"\nrequest_id: {request_id}")


def cli_response(request_id: str) -> None:
    """
    Gets an archive response from the endpoint
    """
    # TODO: Could (with arg?) show a list of the last 5-10 ids to choose
    if request_id == "":
        try:
            with open(".request_id", "r") as f:
                request_id = f.readlines()[-1].strip("\n")
        except Exception:
            print("No request_id provided as argument or in file")

    # Load schema file
    try:
        schema_file = open(os.path.join("build", "schema.json"), "r")
        schema = json.load(schema_file)
    except Exception:
        raise CliError(
            "Cannot load build schema. Try moving to the directory where you deploy your app."
        )

    # Get a token
    fd_url = schema["identity_endpoint"] + "/token"
    token = get_token(fd_url)

    # Make the request
    for app in schema["apps"]:
        app_id = schema["apps"][app]["id"]
    carrier_endpoint = schema["carrier_endpoint"]
    url = f"{carrier_endpoint}/endpoints/{app_id}/response/{request_id}/archive"
    bearer = {"Authorization": f"Bearer {token}", "accept": "application/octet-stream"}
    resp = requests.get(url, headers=bearer, params={"pattern": ".>", "format": "json_array"})
    print(f"\nResponse archive for request_id {request_id}:")
    print(resp.content)


def init(project_name: str) -> None:
    """
    Initializes a project using a template
    """
    cookiecutter_template = "https://github.com/seaplane-io/seaplane-app-python-template.git"
    project_directory = "."

    if not project_name.isidentifier():
        suggestion = re.sub("[^A-Za-z0-9_]", "", project_name)
        raise CliError(f'can\'t make a new project named "{project_name}". Try "{suggestion}"')

    extra_context = {"project_slug": project_name, "seaplane_version": version("seaplane")}

    cookiecutter(
        cookiecutter_template,
        output_dir=project_directory,
        no_input=True,  # Disable any interactive prompts
        extra_context=extra_context,
    )

    print(f"🛩️  {project_name} project generated successfully!")


def main() -> None:
    parser = ArgumentParser(prog="seaplane", description="Seaplane Apps command line interface")
    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    # Build command
    subparsers.add_parser("build", help="Build command")

    # Deploy command
    subparsers.add_parser("deploy", help="Deploy command")

    # Destroy command
    subparsers.add_parser("destroy", help="Remove your App and associated tasks")

    # Init command
    init_parser = subparsers.add_parser("init", help="Init command")
    init_parser.add_argument("app", help="Seaplane Apps name")

    # Status command
    stat_parser = subparsers.add_parser("status", help="Status of your app and resources")
    stat_parser.add_argument(
        "--watch",
        "-w",
        default=False,
        action="store_true",
        help="Use UNIX watch command to refresh status every 2s",
    )

    # Endpoints commands
    req_parser = subparsers.add_parser("request", help="Send data to the request endpoint")
    req_parser.add_argument(
        "--data", "-d", type=str, help="Request body, @ to load a file, @- for stdin"
    )
    resp_parser = subparsers.add_parser("response", help="Get data from the response endpoint")
    resp_parser.add_argument("--request_id", "-r", type=str, help="Request ID")

    # Version argument
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=version("seaplane")),
    )

    args = parser.parse_args()

    try:
        if args.command == "build":
            cli_build()

        elif args.command == "deploy":
            cli_deploy()

        elif args.command == "destroy":
            cli_destroy()

        elif args.command == "init":
            init(args.app)

        elif args.command == "status":
            cli_status(args.watch)

        elif args.command == "request":
            if args.data is None:
                raise CliError("Error: expected --data argument to be provided")

            cli_request(args.data)

        elif args.command == "response":
            request_id = ""
            if args.request_id:
                request_id = args.request_id
            cli_response(request_id)

        else:
            parser.print_help()

    except CliError as e:
        print(e)
        exit(1)

    exit(0)


if __name__ == "__main__":
    main()



from typing import Any, Generator, List, Optional
from importlib.metadata import version
import json
import os
import sys

import toml

from seaplane.apps.debug_schema import build_debug_schema
from seaplane.apps.decorators import legacy_context
from seaplane.apps.status import status
from seaplane.logs import log
from seaplane.pipes.executor import execute, Message
from seaplane.deploy import deploy, destroy

# importing this because it loads our .env file as a side effect
from seaplane.config import config  # noqa: F401


class LegacyTaskContextAdapter:
    """
    LegacyTaskContextAdapter is given to tasks running in legacy mode.
    """

    def __init__(self, parent: Message):
        self.parent = parent
        self.queue: List[Optional[bytes]] = []

    def emit(self, message: bytes, batch_id: Any = None) -> None:
        """Queues message to send once the task completes."""
        if batch_id is not None:
            raise RuntimeError(
                "Explicit batch ids are no longer supported."
                " Batch ids will be associated automatically with output"
            )

        self.queue.append(message)

    def cont(self) -> None:
        """Continues a task without emitting anything."""
        self.queue.append(None)

    @property
    def request_id(self) -> str:
        return self.parent.request_id

    @property
    def object_data(self) -> bytes:
        return self.parent.object_data

    @property
    def body(self) -> bytes:
        return self.parent.body


def _start_task(task_id: str) -> None:
    log.logger.info(f"executing task in legacy mode: {task_id}")

    flow = None
    for compat in legacy_context.apps:
        compat.assemble()

    for compat in legacy_context.apps:
        dag = compat.real_dag

        if task_id in dag.flow_registry:
            flow = dag.flow_registry[task_id]
            break

    if flow is None:
        raise RuntimeError(f"no task found with id {task_id}")

    def legacy_task_work(root_context: Message) -> Generator[Any, None, None]:
        tc_adapter = LegacyTaskContextAdapter(root_context)
        flow.work(tc_adapter)

        for message in tc_adapter.queue:
            yield message

    execute(flow.instance_name, legacy_task_work)


def start() -> None:
    """
    This is a legacy entry point. Prefer "Dag.run()" for new developments.

    *Run Locally*, this function builds, deploys, and manages a customer application.
    Run on seaplane infrastructure with no arguments, this function
    runs a user @task in a loop, pulling information upstream from the seaplane platform.
    """

    log.info(f"\n\n\tSeaplane Apps version {version('seaplane')}, Legacy 0.5 mode\n")

    if len(sys.argv) > 1:
        command = sys.argv[1]

        if command == "build":
            for compat in legacy_context.apps:
                compat.assemble()

            debug_schema = build_debug_schema([compat.real_dag for compat in legacy_context.apps])
            print(json.dumps(debug_schema, indent=2))

        elif command == "deploy":
            pyproject = toml.loads(open("pyproject.toml", "r").read())
            project_directory_name = pyproject["tool"]["poetry"]["name"]
            for compat in legacy_context.apps:
                compat.assemble()

            for compat in legacy_context.apps:
                deploy(compat.real_app, project_directory_name)

            legacy_schema = build_debug_schema([compat.real_dag for compat in legacy_context.apps])
            with open(os.path.join("build", "schema.json"), "w") as file:
                json.dump(legacy_schema, file, indent=2)

        elif command == "destroy":
            for compat in legacy_context.apps:
                compat.assemble()

            for compat in legacy_context.apps:
                destroy(compat.real_app)

        elif command == "status":
            status()
        else:
            log.error(
                f"Found an invalid internal command `{command}`.\n"
                + "Expected one of: \n"
                + " - build\n"
                + " - deploy\n"
                + " - destroy\n"
                + " - status\n"
            )
        return None

    task_id = os.getenv("TASK_ID") or os.getenv("INSTANCE_NAME")

    if not task_id:
        log.debug("ERROR: Could not find TASK_ID or INSTANCE_NAME value")
        log.error(
            "Executing a Smartpipe workflow from outside a "
            + "Seaplane Deployment is not currently supported!"
        )
        sys.exit(-1)

    log.info(f"Starting Task {task_id} ...")
    _start_task(task_id)
    return None



from typing import cast

from seaplane_framework.api.exceptions import ApiException

from seaplane.errors import SeaplaneError
from seaplane.logs import log
from seaplane.object import object_store


def create_bucket_if_needed(app_id: str, bucket_name: str) -> str:
    """
    Returns the notification stream associated with a given bucket,
    after either creating the bucket with a default notification subject name,
    or reading the notification subject name from an existing bucket.

    Will throw an exception if the bucket cannot be created, or if the
    bucket already exists but is not configured to send notifications.
    """
    try:
        bkt = object_store.get_bucket(bucket_name)
        if "notify" not in bkt:
            raise SeaplaneError(
                f"bucket {bucket_name} already exists, but is not configured for notifications."
                " The seaplane project can create a new correctly configured bucket for you,"
                " but if you want to create one yourself, make sure it has an associated"
                " notify subject for your task to listen to."
            )
        ret = cast(str, bkt["notify"])
        log.debug(f"using existing bucket {bucket_name} with subject {ret}")
        return ret
    except ApiException as e:
        if e.status != 404:
            raise

    log.info(f"creating bucket {bucket_name} with default attributes")
    notify = f"{bucket_name}-evts.updates"
    bucket_def = {
        "description": f"Automatically created for application {app_id}",
        "replicas": 3,
        "max_bytes": -1,
        "notify": notify,
    }
    object_store.create_bucket(bucket_name, bucket_def)

    return notify
